"""Example of customizing evaluation with RLlib.

Pass --custom-eval to run with a custom evaluation function too.

Here we define a custom evaluation method that runs a specific sweep of env
parameters (SimpleCorridor corridor lengths).

------------------------------------------------------------------------
Sample output for `python custom_eval.py`
------------------------------------------------------------------------

INFO algorithm.py:623 -- Evaluating current policy for 10 episodes.
INFO algorithm.py:650 -- Running round 0 of parallel evaluation (2/10 episodes)
INFO algorithm.py:650 -- Running round 1 of parallel evaluation (4/10 episodes)
INFO algorithm.py:650 -- Running round 2 of parallel evaluation (6/10 episodes)
INFO algorithm.py:650 -- Running round 3 of parallel evaluation (8/10 episodes)
INFO algorithm.py:650 -- Running round 4 of parallel evaluation (10/10 episodes)

Result for PG_SimpleCorridor_2c6b27dc:
  ...
  evaluation:
    custom_metrics: {}
    episode_len_mean: 15.864661654135338
    episode_reward_max: 1.0
    episode_reward_mean: 0.49624060150375937
    episode_reward_min: 0.0
    episodes_this_iter: 133
    off_policy_estimator: {}
    policy_reward_max: {}
    policy_reward_mean: {}
    policy_reward_min: {}
    sampler_perf:
      mean_env_wait_ms: 0.0362923321333299
      mean_inference_ms: 0.6319202064080927
      mean_processing_ms: 0.14143652169068222

------------------------------------------------------------------------
Sample output for `python custom_eval.py --custom-eval`
------------------------------------------------------------------------

INFO algorithm.py:631 -- Running custom eval function <function ...>
Update corridor length to 4
Update corridor length to 7
Custom evaluation round 1
Custom evaluation round 2
Custom evaluation round 3
Custom evaluation round 4

Result for PG_SimpleCorridor_0de4e686:
  ...
  evaluation:
    custom_metrics: {}
    episode_len_mean: 9.15695067264574
    episode_reward_max: 1.0
    episode_reward_mean: 0.9596412556053812
    episode_reward_min: 0.0
    episodes_this_iter: 223
    foo: 1
    off_policy_estimator: {}
    policy_reward_max: {}
    policy_reward_mean: {}
    policy_reward_min: {}
    sampler_perf:
      mean_env_wait_ms: 0.03423667269562796
      mean_inference_ms: 0.5654563161491506
      mean_processing_ms: 0.14494765630060774
"""

import argparse
import functools
import os

import ray
from ray import air, tune
from ray.rllib.algorithms.ppo import PPOConfig
from ray.rllib.evaluation.metrics import collect_episodes, summarize_episodes
from ray.rllib.examples.env.simple_corridor import SimpleCorridor
from ray.rllib.utils.test_utils import check_learning_achieved

parser = argparse.ArgumentParser()
parser.add_argument("--evaluation-parallel-to-training", action="store_true")
parser.add_argument("--num-cpus", type=int, default=0)
parser.add_argument(
    "--framework",
    choices=["tf", "tf2", "torch"],
    default="torch",
    help="The DL framework specifier.",
)
parser.add_argument("--async-eval", action="store_true")
parser.add_argument("--no-custom-eval", action="store_true")
parser.add_argument(
    "--as-test",
    action="store_true",
    help="Whether this script should be run as a test: --stop-reward must "
    "be achieved within --stop-timesteps AND --stop-iters.",
)
parser.add_argument(
    "--stop-iters", type=int, default=50, help="Number of iterations to train."
)
parser.add_argument(
    "--stop-timesteps", type=int, default=20000, help="Number of timesteps to train."
)
parser.add_argument(
    "--stop-reward", type=float, default=0.7, help="Reward at which we stop training."
)
parser.add_argument(
    "--local-mode",
    action="store_true",
    help="Init Ray in local mode for easier debugging.",
)


def custom_eval_function(algorithm, eval_workers):
    """Example of a custom evaluation function.

    Args:
        algorithm: Algorithm class to evaluate.
        eval_workers: Evaluation WorkerSet.

    Returns:
        metrics: Evaluation metrics dict.
    """
    # Set different env settings for each worker. Here we use the worker's
    # `worker_index` property.
    eval_workers.foreach_worker(
        func=lambda w: w.foreach_env(
            lambda env: env.set_corridor_length(4 if w.worker_index == 1 else 7)
        )
    )

    for i in range(5):
        print("Custom evaluation round", i)
        # Calling .sample() runs exactly one episode per worker due to how the
        # eval workers are configured.
        eval_workers.foreach_worker(func=lambda w: w.sample(), local_worker=False)

    # Collect the accumulated episodes on the workers, and then summarize the
    # episode stats into a metrics dict.
    episodes = collect_episodes(workers=eval_workers, timeout_seconds=99999)
    # You can compute metrics from the episodes manually, or use the
    # convenient `summarize_episodes()` utility:
    metrics = summarize_episodes(episodes)

    # You can also put custom values in the metrics dict.
    metrics["foo"] = 1
    return metrics


def custom_async_eval_function(algorithm, eval_workers, weigths_ref, weights_seq_no):
    """Example of a custom asynchronous evaluation function.

    Args:
        algorithm: Algorithm class to evaluate.
        eval_workers: Evaluation WorkerSet.
        weights_ref: Object reference (`ObjectRef`) to the module weights in Ray's
            object store.
        weights_seq_no: Integer, identifying the synchronization round. This
            identifies during ayschnronous evaluation, if the weights are already
            synched between evaluation workers and learners.

    Returns:
        metrics: Evaluation metrics dict.
    """

    def remote_fn(worker, weights_ref, weights_seq_no):

        # Pass in weights seq-no so that eval workers may ignore this call if no update
        # has happened since the last call to `remote_fn` (sample).
        worker.set_weights(weights=ray.get(weights_ref), weights_seq_no=weights_seq_no)

        # Set different env settings for each worker. Here we use the worker's
        # `worker_index` property.
        for env in worker.env.envs:
            env.set_corridor_length(4 if worker.worker_index == 1 else 7)

        # Sample 5 episodes.
        episodes = worker.sample(explore=False, num_episodes=5)
        # Get the metrics.
        metrics = worker.get_metrics()

        return worker.worker_index, episodes, metrics, weights_seq_no

    # Create a partial to bake in the `weights_ref.` such that we can delete later this
    # partial and resolve any pointers to `weigths_ref` in Ray's object store. Note,
    # this avoids a memory leak.
    partial_remote_fn = functools.partial(
        remote_fn,
        weigths_ref=weigths_ref,
        weights_seq_no=weights_seq_no,
    )

    # Sample from all remote workers asynchronously.
    eval_workers.foreach_worker_async(
        func=partial_remote_fn,
        healthy_only=True,
    )

    eval_results = eval_workers.fetch_ready_async_reqs()

    episodes = []
    rollout_metrics = []
    for _, result in eval_results:
        _, eps, metrics, seq_no = result

        # Ignore results, if the weights seq-number does not match (is
        # from a previous evaluation step) OR if we have already reached
        # the configured duration (e.g. number of episodes to evaluate
        # for).
        if seq_no == algorithm._evaluation_weights_seq_no:
            episodes.extend(eps)
            rollout_metrics.extend(metrics)

    # Remove the pointer to the weight reference.
    del partial_remote_fn

    sampler_results = summarize_episodes(rollout_metrics)
    metrics = dict({"sampler_results": sampler_results})

    return metrics


if __name__ == "__main__":
    args = parser.parse_args()

    if args.no_custom_eval:
        eval_fn = None
        async_eval_fn = None
    else:
        if args.async_eval:
            async_eval_fn = custom_async_eval_function
            eval_fn = None
        else:
            async_eval_fn = None
            eval_fn = custom_eval_function

    ray.init(num_cpus=args.num_cpus or None, local_mode=args.local_mode)

    config = (
        PPOConfig()
        .environment(SimpleCorridor, env_config={"corridor_length": 10})
        # Training rollouts will be collected using just the learner
        # process, but evaluation will be done in parallel with two
        # workers. Hence, this run will use 3 CPUs total (1 for the
        # learner + 2 more for evaluation workers).
        .rollouts(num_rollout_workers=0)
        .evaluation(
            enable_async_evaluation=args.async_eval,
            evaluation_num_workers=2,
            # Enable evaluation, once per training iteration.
            evaluation_interval=1,
            # Run 10 episodes each time evaluation runs (OR "auto" if parallel to
            # training).
            evaluation_duration="auto" if args.evaluation_parallel_to_training else 10,
            # Evaluate parallelly to training.
            evaluation_parallel_to_training=args.evaluation_parallel_to_training,
            evaluation_config=PPOConfig.overrides(
                env_config={
                    # Evaluate using LONGER corridor than trained on.
                    "corridor_length": 5,
                },
            ),
            custom_evaluation_function=eval_fn,
            custom_async_evaluation_function=async_eval_fn,
        )
        .framework(args.framework)
        # Use GPUs iff `RLLIB_NUM_GPUS` env var set to > 0.
        .resources(num_gpus=int(os.environ.get("RLLIB_NUM_GPUS", "0")))
    )

    stop = {
        "training_iteration": args.stop_iters,
        "timesteps_total": args.stop_timesteps,
        "episode_reward_mean": args.stop_reward,
    }

    tuner = tune.Tuner(
        "PPO",
        param_space=config.to_dict(),
        run_config=air.RunConfig(stop=stop, verbose=1),
    )
    results = tuner.fit()

    # Check eval results (from eval workers using the custom function),
    # not results from the regular workers.
    if args.as_test:
        check_learning_achieved(results, args.stop_reward, evaluation=True)
    ray.shutdown()
